/*****************************************************************************
 * AMOS2
 *
 * Author: (c) 1998-2006 Tore Risch, UDBL
 * $RCSfile: Foreign.java,v $
 * $Revision: 1.9 $ $Date: 2006/02/15 13:18:19 $
 * $State: Exp $ $Locker:  $
 *
 * Description: Demonstration and test of foreign functions in Java
 * ===========================================================================
 * $Log: Foreign.java,v $
 * Revision 1.9  2006/02/15 13:18:19  torer
 * Demonstration of opaque objects in Java: jear(Date)->Integer in Java
 *
 ****************************************************************************/

import callin.*;
import callout.*;

/**
  * Example class for testing the callout-interface to foreign Java-functions.
  */
public class Foreign {

  /**
    * Must have default constructor.
    */
  public Foreign() {
    // Put initializations here...
  }

  /**
    * The foreign function.
    * After declaring sqrt() as follows in AMOS2:
    *
    *   create function javaSqrt(real) -> real as
    *   as foreign "JAVA:Foreign/sqrtbf";
    *
    * AMOS2 calls this method when the user executes a call to jsqrt();
    */
  public void sqrtbf(CallContext cxt, Tuple tpl) throws AmosException {

    double x;

    x = tpl.getDoubleElem(0);	// Pick up the argument
    if (x < 0.0) {
      // Don't return any value if sqrt() is undefined.
    }
    else if (Math.abs(x) < 1.0E-7) {
      // One root
      tpl.setElem(1, 0.0);
      cxt.emit(tpl);
    }
    else {
      // Two roots
      tpl.setElem(1, Math.sqrt(x));
      cxt.emit(tpl);
      tpl.setElem(1, -Math.sqrt(x));
      cxt.emit(tpl);
    }
}

/**
   The following is an example of the implementation of a multidirectional foreign
   function in Java. It is callable after the following AMOSQL definition:
   create function myabs(number n)-> number a 
          as multidirectional ("bf" foreign "JAVA:Foreign/absbf")
                              ("fb" foreign "JAVA:Foreign/absfb");
*/

   public void absbf(CallContext cxt, Tuple tpl) throws AmosException {

     double x;
     
     x = tpl.getDoubleElem(0); // pick up first argument
     if(x<0) tpl.setElem(1,-x);
     else    tpl.setElem(1,x);
     cxt.emit(tpl);
}

   public void absfb(CallContext cxt, Tuple tpl) throws AmosException {
     // Inverse of abs(x)
     double x;
     
     x = tpl.getDoubleElem(1); // pick up result
     if(x==0.0)
     {
       tpl.setElem(0,x);
       cxt.emit(tpl);
     }
     else
     {
       tpl.setElem(0,x);
       cxt.emit(tpl);
       tpl.setElem(0,-x);
       cxt.emit(tpl);
     }
}
    public void bigstring(CallContext cxt, Tuple tpl) throws AmosException
    {
	int sz = tpl.getIntElem(0),i;
        StringBuffer b=new StringBuffer();

        for(i=0;i<sz;i++)
	    b = b.append("!");
        tpl.setElem(1,new String(b));
        cxt.emit(tpl);
    }

  public void jotabf(CallContext cxt, Tuple tpl) throws AmosException
  {

     int low=tpl.getIntElem(0), up=tpl.getIntElem(1), i;

     for(i=low;i<=up;i++)
     {
        tpl.setElem(2,i);
        cxt.emit(tpl);
     }
  }
    public void buggy(CallContext cxt, Tuple tpl) throws AmosException
    {
        double i=1.9, j=i/0.0;
    }
    /* To crash the JVM do this:
       create function buggy()-> Boolean as
       foreign "JAVA:Foreign/buggy";
       buggy(); */

  public void jearbf(CallContext cxt, Tuple tpl) throws AmosException
    { /* Get the year out of a date. 
         Demos passing of opaque objects as type Oid. 
         create function jear(Date d)->Integer 
                as foreign "JAVA:Foreign/jearbf";
         Test: jear(date(now()));
       */
        Scan s;
        Tuple arg = new Tuple(1), res;
        Connection theConnection = new Connection("");
        
        arg.setElem(0,tpl.getElem(0));
        s = theConnection.callFunction("YEAR",arg,1);
        res = s.getRow();
        if(res == null) return;
        tpl.setElem(1,res.getIntElem(0));
        cxt.emit(tpl);
    } 
}


